// Title: Mad Bulb
// Based on: Mandelbulb Deconstructed
// Original author: morgan3d
// URL: https://www.shadertoy.com/view/XsXXWS
// License: https://opensource.org/licenses/BSD-2-Clause
// This modified version was used in the Xmas Demo 2021 by Julin & Corneliusen: https://www.ignorantus.com/pages/xmasdemo2021/

// Copyright 2014 Morgan McGuire
// The 2-Clause BSD License
// Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
// conditions are met:
// 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
// disclaimer in the documentation and/or other materials provided with the distribution.
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
// BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
// SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#version 310 es
precision highp float;

//#define XAVIER

in vec2 v_UV1;
uniform float u_Dissolve;

#define v_texCoord v_UV1

layout(location = 0) out vec4 outColor;

layout(std140, binding = 0) uniform ctx_stuff {
    unsigned int width;
    unsigned int height;
    float aspect;
    float time;
    unsigned int frame;
};

layout(std140, binding = 2) uniform GPUMadBulb
{
	float starttime;
	float pitch;
	float yaw;
	float zoom;
	float power;
};

#define iTime (time-starttime)

#define PI 3.1415926535f

#define SLCOLS 7
vec3 slideCol( vec3 v, float h )
{
	vec3 cols[SLCOLS];
	cols[0] = v.rgb;
	cols[1] = v.rgb;
	cols[2] = v.brg;
	cols[3] = v.rgb;
	cols[4] = v.rgb;
	cols[5] = v.gbr;
	cols[6] = cols[0];

	h = h - floor( h );
	float Slice     = float(SLCOLS-1) * h;
	float SliceInt  = floor( Slice );
	float SliceFrac = Slice - SliceInt;

	int i = int(SliceInt);
	return mix( cols[i], cols[i + 1], SliceFrac );
}

// satan2() function from the book Real Programming (ISBN 9781678050443) by Corneliusen & Julin
// https://www.ignorantus.com/books/real/
// This is still a lot faster than using atan(y,x) since it is reordered.
#define f2u( x ) floatBitsToUint( x )
#define u2f( x ) uintBitsToFloat( x )

float satan2( float y, float x )
{
    uint sign_mask = 0x80000000u;
    float b = 0.596227f;

    // Extract the sign bits
    uint ux_s  = (sign_mask & f2u(x) )^sign_mask;
    uint uy_s  = (sign_mask & f2u(y) )^sign_mask;

    // Determine the quadrant offset
    float q = float( ( ~ux_s & uy_s ) >> 29 | ux_s >> 30 );

    // Calculate the arctangent in the first quadrant
    float bxy_a = abs( b * x * y );
    float num = bxy_a + y * y;
    float atan_1q =  num / ( x * x + bxy_a + num );

    // Translate it to the proper quadrant
    uint uatan_2q = ((ux_s ^ uy_s) | f2u(atan_1q));
    return (q + u2f(uatan_2q)) * PI/2.0f - PI;
}

////////////////////////////////////////////////////////////
// Scene configuration:

// = normalize(+1, +1, -1)
const vec3 lightDirection = vec3(0.57735026919, 0.57735026919, -0.57735026919);

const vec3 keyLightColor  = vec3(1.0, 1.0, 1.0);

const float cameraDistance = 5.0;


#ifdef XAVIER
#define MIN_DIST 0.003f
#define FRAC_ITER 10
#define MARCH_ITER 100
#else
#define MIN_DIST 0.0003f
#define FRAC_ITER 16
#define MARCH_ITER 150
#endif

// A small step, used for computing the surface normal
// by numerical differentiation. A scaled up version of
// this is also used for computing a low-frequency gradient.
const vec3 eps = vec3(MIN_DIST * 5.0, 0.0, 0.0);

// Orientation of the object
mat3 rotation;

////////////////////////////////////////////////////////////

float getAO( vec3 P )
{
	P = rotation * P;
	
	const float externalBoundingRadius = 1.2f;
	if( length(P) - externalBoundingRadius > 1.0f ) return 1.0f;

	vec3 Q = P;

	float AO = 1.0f;

	// Used to smooth discrete iterations into continuous distance field
	// (similar to the trick used for coloring the Mandelbrot set)	
	float derivative = 1.0;
	
	int i;
	for( i = 0; i < FRAC_ITER; ++i ) {

		// Darken as we go deeper
		AO *= 0.725;

		float r = length(Q);
		if (r > 2.0) break;

		// Convert to polar coordinates and then rotate by the power
		float theta = acos(Q.z / r) * power;
		float phi   = satan2(Q.y, Q.x) * power;
			
		// Update the derivative
		derivative = pow(r, power - 1.0) * power * derivative + 1.0;
			
		// Convert back to Cartesian coordinates and 
		// offset by the original point (which we're orbiting)
		float sinTheta = sin(theta);
			
		Q = vec3(sinTheta * cos(phi), sinTheta * sin(phi), cos(theta)) * pow(r, power) + P;
	}

	if( i == FRAC_ITER ) return AO;

	// The point escaped. Remap AO for more brightness and return
	return min((AO + 0.075) * 4.1, 1.0);
}

float distanceToSurface( vec3 P )
{
	P = rotation * P;
	
	const float externalBoundingRadius = 1.2;
	if( length(P) - externalBoundingRadius > 1.0f ) return 1.0f;

	vec3 Q = P;

	// Used to smooth discrete iterations into continuous distance field
	// (similar to the trick used for coloring the Mandelbrot set)	
	float derivative = 1.0;

	int i;
	float r;
	float theta, phi, sinTheta;
	for( i = 0; i < FRAC_ITER; i++ ) {

		r = length(Q);
		if( r > 2.0f ) break;

		// Convert to polar coordinates and then rotate by the power
		theta = acos(Q.z / r) * power;
		phi   = satan2(Q.y, Q.x) * power;
			
		// Update the derivative
		derivative = pow(r, power - 1.0) * power * derivative + 1.0;
			
		// Convert back to Cartesian coordinates and 
		// offset by the original point (which we're orbiting)
		sinTheta = sin(theta);
			
		Q = vec3(sinTheta * cos(phi), sinTheta * sin(phi), cos(theta)) * pow(r, power) + P;

	}

	// Never escaped, so either already in the set...or a complete miss
	if( i == FRAC_ITER ) return MIN_DIST;

	// Embed a sphere within the fractal to fill in holes under low iteration counts
	const float internalBoundingRadius = 0.72;

	// The point escaped.
	return min(length(P) - internalBoundingRadius, 0.5 * log(r) * r / derivative);
}

vec3 trace(vec2 coord) {

	vec3 rayOrigin = vec3( 2.0f * coord - 1.0f, -cameraDistance );

	// Correct for aspect ratio
	rayOrigin.x *= aspect;
	
	vec3 rayDirection = normalize(normalize(vec3(0.0, 0.0, 1.0) - rayOrigin) + 0.2 * vec3(rayOrigin.xy, 0.0) / zoom);

	// Distance from ray origin to hit point
	float t = 0.0;
	
	// Point on (technically, near) the surface of the Mandelbulb
	vec3 X;

	float d;

	// March along the ray, detecting when we are very close to the surface
	int i;
	for (i = 0; i < MARCH_ITER; ++i) {
		X = rayOrigin + rayDirection * t;
		
		d = distanceToSurface(X);
		if( d < MIN_DIST ) break;
		
		// Advance along the ray by the worst-case distance to the
		// surface in any direction
		t += d;
	}

	if( i == MARCH_ITER ) return vec3( -1.0f );

	vec3 color;

	float AO = getAO( X );
				
	// Back away from the surface a bit before computing the gradient
	X -= rayDirection * eps.x;
				
	// Accurate micro-normal
	vec3 n = normalize( vec3(d - distanceToSurface(X - eps.xyz),
							 d - distanceToSurface(X - eps.yxz),
							 d - distanceToSurface(X - eps.zyx)));
		
	// Broad scale normal to large shape
	vec3 n2 = normalize( vec3(d - distanceToSurface(X - eps.xyz * 50.0),
							  d - distanceToSurface(X - eps.yxz * 50.0),
							  d - distanceToSurface(X - eps.zyx * 50.0)));

	// Acid colors instead of the icy crap on ShaderToy
	vec3 flc = slideCol( n2, iTime/18.0f )*1.33f;
	return AO * mix(flc, keyLightColor, AO * clamp(0.7 * dot(lightDirection, n) + 0.6, 0.0, 1.0));
}

#ifdef XAVIER
vec4 getBG( void )
{
	vec3 plane_col = vec3( 0.5f, 0.0f, 0.0f );
	vec2 dxy = vec2( (v_texCoord.x - 0.5f)*16.0f/9.0f, v_texCoord.y - 0.5f );
	float dv = (1.0f - sqrt( dot( dxy, dxy ) + (sin(fract(time/2.0f)*2.0f*3.14f)+1.0f)*0.5f ))*2.0f;
	return vec4( plane_col * dv, 1.0f - u_Dissolve );
}
#endif

void main()
{
	vec2 co = vec2( v_UV1.x, 1.0f - v_UV1.y );

	// Euler-angle animated rotation	
	rotation    = mat3(1.0, 0.0, 0.0, 0.0, cos(pitch), -sin(pitch), 0.0, sin(pitch), cos(pitch)) *
	              mat3(cos(yaw), 0.0, sin(yaw), 0.0, 1.0, 0.0, -sin(yaw), 0.0, cos(yaw));

	vec3 color = trace(co);

	if( color.r == -1.0f ) {
#ifdef XAVIER
		outColor  = getBG();
#else
		outColor = vec4( 0.0f );
#endif
		return;
	}

	// Coarse RGB->sRGB encoding via sqrt
	color = sqrt(color);
	
    outColor = vec4(color, 1.0 - u_Dissolve);
}
